
/** @module : Mem_System_Wrapper
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2010 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
// Mem_System_Wrapperfor local memory and network 
module mem_system_wrapper #(parameter CORE = 0, ID_BITS = 4, REAL_ADDR_BITS = 16, 
							DATA_WIDTH = 32, INDEX_BITS = 6, OFFSET_BITS = 3, 
							ADDRESS_BITS = 32, INIT_FILE = "memory.mem",IN_PORTS = 1, 
							OUT_PORTS = 1, VC_BITS = 1, VC_DEPTH_BITS = 4,
                            EXTRA  = 2, TYPE_BITS = 2)( // report is used to show  
						                                // performance statistics
             		 clock, reset,
				     i_read, i_write, i_address, i_in_data, 
				     i_out_addr, i_out_data, i_valid, i_ready,
				     
				     d_read, d_write, d_address, d_in_data, 
				     d_out_addr, d_out_data, d_valid, d_ready, 
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full,
					 // performance reporting
					 report
);										   

localparam VC_PER_PORTS  = (1 << VC_BITS); 
localparam FLOW_BITS = (2*ID_BITS) + EXTRA;
localparam FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + DATA_WIDTH;
localparam NUM_CACHES = 2,  MSG_BITS = 3;
localparam M_CACHES = NUM_CACHES + 1; // one for the network interface 

localparam NO_REQ = 0, WB_REQ = 1, C_SENT = 2, C_RECV = 3, R_REQ = 4, W_REQ = 5;

input clock, reset;
input  i_read, i_write;
input  [ADDRESS_BITS-1:0] i_address;
input  [DATA_WIDTH-1:0] i_in_data;

output i_valid, i_ready;
output [DATA_WIDTH-1:0] i_out_data;
output [ADDRESS_BITS-1:0] i_out_addr;

input  d_read, d_write;
input  [ADDRESS_BITS-1:0] d_address;
input  [DATA_WIDTH-1:0] d_in_data;

output d_valid, d_ready;
output [DATA_WIDTH-1:0] d_out_data;
output [ADDRESS_BITS-1:0] d_out_addr;

output [(FLIT_WIDTH * IN_PORTS) -1:0] from_core_flit;
output [IN_PORTS-1: 0] v_from_core;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;
input  report;
 
wire [1:0] read = {d_read, i_read}; 
wire [1:0] write = {d_write, i_write};
wire [(2*ADDRESS_BITS)-1:0] address = {d_address, i_address};
wire [(2*DATA_WIDTH)-1:0] in_data   = {d_in_data, i_in_data};

wire [1:0] valid, ready;
wire[(2*DATA_WIDTH)-1:0] out_data;
wire[(2*ADDRESS_BITS)-1:0] out_address;

assign i_valid = valid[0]; 
assign i_ready = ready [0];
assign i_out_data = out_data[DATA_WIDTH-1 :0];
assign i_out_addr = out_address[ADDRESS_BITS-1 :0];

assign d_valid = valid[1]; 
assign d_ready = ready [1];
assign d_out_data = out_data[(2*DATA_WIDTH)-1 -:DATA_WIDTH];
assign d_out_addr = out_address[(2*ADDRESS_BITS)-1 -:ADDRESS_BITS];

wire  [(ID_BITS - 1): 0] Core_ID = CORE; 

wire  [(MSG_BITS * NUM_CACHES)-1:0]    cache2mem_msg; 
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] cache2mem_address; 
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] initial_cache2mem_addr; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   cache2mem_data;

wire [(MSG_BITS * M_CACHES)-1:0]       lm_mem2cache_msg; 
wire [(ADDRESS_BITS * M_CACHES)-1:0]   lm_mem2cache_address;
wire [(DATA_WIDTH * M_CACHES)-1:0]     lm_mem2cache_data;

wire [MSG_BITS-1:0]     network2mem_msg; 
wire [ADDRESS_BITS-1:0] network2mem_address; 
wire [DATA_WIDTH-1:0]   network2mem_data;

wire [(MSG_BITS * NUM_CACHES)-1:0]     network2cache_msg; 
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] network2cache_address; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   network2cache_data;

wire [MSG_BITS-1:0]      o_cache2mem_msg      = 0; 
wire [ADDRESS_BITS-1:0]  o_cache2mem_address  = 0; 
wire [DATA_WIDTH-1:0]    o_cache2mem_data     = 0; 
wire [ID_BITS - 1: 0]    id_pad               = 0; 

wire [ADDRESS_BITS-1:0] i_initial_addr = initial_cache2mem_addr[(ADDRESS_BITS-1) -: ADDRESS_BITS]; 
wire [ADDRESS_BITS-1:0] d_initial_addr = initial_cache2mem_addr[((2*ADDRESS_BITS)-1) -: ADDRESS_BITS];

reg  [(ID_BITS - 1): 0] active_dest; 
wire [ID_BITS-1:0]      i_addr_dest    = i_initial_addr >> REAL_ADDR_BITS;
wire [(ID_BITS - 1): 0] i_destination  = reset? 0 : (i_addr_dest > 0)? 
													 i_addr_dest : active_dest; 
													// for the current version, instruction 
                                                    // cannot go from remote to local
wire  i_local_addr = (i_destination == Core_ID); 
wire  [(ID_BITS - 1): 0] d_destination = d_initial_addr >> REAL_ADDR_BITS;
wire  d_local_addr = (d_destination == Core_ID);
													
wire [ADDRESS_BITS-1:0] i_adjusted_address  = {i_destination, i_initial_addr[REAL_ADDR_BITS-1:0]}; 
wire [ADDRESS_BITS-1:0] d_adjusted_address  = d_initial_addr; 

assign cache2mem_address = {d_adjusted_address, i_adjusted_address};
wire [ADDRESS_BITS-1:0]  iFull_cache2mem_address    = cache2mem_address[(ADDRESS_BITS-1) -: ADDRESS_BITS]; 
wire [ADDRESS_BITS-1:0]  dFull_cache2mem_address    = cache2mem_address[((2*ADDRESS_BITS)-1) -: ADDRESS_BITS];
wire [REAL_ADDR_BITS-1:0]  iShort_cache2mem_address = iFull_cache2mem_address; 
wire [REAL_ADDR_BITS-1:0]  dShort_cache2mem_address = dFull_cache2mem_address;

wire [MSG_BITS-1:0]      i_cache2mem_msg     = cache2mem_msg[(MSG_BITS-1) -: MSG_BITS];  
wire [ADDRESS_BITS-1:0]  i_cache2mem_address = i_local_addr ? {id_pad,iShort_cache2mem_address}: iFull_cache2mem_address; 
wire [DATA_WIDTH-1:0]    i_cache2mem_data    = cache2mem_data[(DATA_WIDTH-1) -: DATA_WIDTH]; 
	    
wire [MSG_BITS-1:0]      d_cache2mem_msg     = cache2mem_msg[((2*MSG_BITS)-1) -: MSG_BITS]; 
wire [ADDRESS_BITS-1:0]  d_cache2mem_address = d_local_addr ?  {id_pad, dShort_cache2mem_address} : dFull_cache2mem_address;
wire [DATA_WIDTH-1:0]    d_cache2mem_data    = cache2mem_data[((2*DATA_WIDTH)-1) -: DATA_WIDTH];

wire [(MSG_BITS * NUM_CACHES)-1:0]     m_cache2mem_msg = i_local_addr ? d_local_addr? 
														 cache2mem_msg : {o_cache2mem_msg, i_cache2mem_msg}
														 : d_local_addr? {d_cache2mem_msg, o_cache2mem_msg}
														 : 0;
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] m_cache2mem_address = i_local_addr ? d_local_addr? 
														     {d_cache2mem_address, i_cache2mem_address} : 
														     {o_cache2mem_address, i_cache2mem_address}
														     : d_local_addr? {d_cache2mem_address, o_cache2mem_address}
														     : 0; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   m_cache2mem_data = i_local_addr ? d_local_addr? 
														 cache2mem_data : {o_cache2mem_data, i_cache2mem_data}
														 : d_local_addr? {d_cache2mem_data, o_cache2mem_data}
														 : 0;
														 
wire [(MSG_BITS * NUM_CACHES)-1:0]     n_cache2mem_msg = ~i_local_addr ? ~d_local_addr? 
														 cache2mem_msg : {o_cache2mem_msg, i_cache2mem_msg}
														 : ~d_local_addr? {d_cache2mem_msg, o_cache2mem_msg}
														 : 0;
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] n_cache2mem_address = ~i_local_addr ? ~d_local_addr? 
														     cache2mem_address : {o_cache2mem_address, i_cache2mem_address}
														     : ~d_local_addr? {d_cache2mem_address, o_cache2mem_address}
														     : 0; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   n_cache2mem_data = ~i_local_addr ? ~d_local_addr? 
														 cache2mem_data : {o_cache2mem_data, i_cache2mem_data}
														 : ~d_local_addr? {d_cache2mem_data, o_cache2mem_data}
														 : 0;

wire  [(MSG_BITS * M_CACHES)-1:0]      lm_cache2mem_msg      = {network2mem_msg, m_cache2mem_msg}; 
wire [(ADDRESS_BITS * M_CACHES)-1:0]   lm_cache2mem_address  = {network2mem_address, m_cache2mem_address}; 
wire [(DATA_WIDTH * M_CACHES)-1:0]     lm_cache2mem_data     = {network2mem_data, m_cache2mem_data};

wire [MSG_BITS-1:0]     mem2network_msg     = lm_mem2cache_msg [(MSG_BITS * M_CACHES)-1 -:MSG_BITS];
wire [ADDRESS_BITS-1:0] mem2network_address = lm_mem2cache_address [(ADDRESS_BITS * M_CACHES)-1 -:ADDRESS_BITS];
wire [DATA_WIDTH-1:0]   mem2network_data    = lm_mem2cache_data [(DATA_WIDTH * M_CACHES)-1 -:DATA_WIDTH];

wire [(MSG_BITS * NUM_CACHES)-1:0]     cache2network_msg      = n_cache2mem_msg; 
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] cache2network_address  = n_cache2mem_address; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   cache2network_data     = n_cache2mem_data;
																								 
wire [(MSG_BITS * NUM_CACHES)-1:0]     m_mem2cache_msg      =  lm_mem2cache_msg [(MSG_BITS * NUM_CACHES)-1:0]; 
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] m_mem2cache_address  =  lm_mem2cache_address [(ADDRESS_BITS * NUM_CACHES)-1:0]; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   m_mem2cache_data     =  lm_mem2cache_data [(DATA_WIDTH * NUM_CACHES)-1:0];

wire [ADDRESS_BITS-1:0]  imem_address    = m_mem2cache_address[(ADDRESS_BITS-1) -: ADDRESS_BITS]; 
wire [ADDRESS_BITS-1:0]  dmem_address    = m_mem2cache_address[((2*ADDRESS_BITS)-1) -: ADDRESS_BITS];
wire [ADDRESS_BITS-1:0]  reform_i_mem2cache_addr = {i_addr_dest, imem_address[REAL_ADDR_BITS-1:0]}; 
wire [ADDRESS_BITS-1:0]  reform_d_mem2cache_addr = {d_destination, dmem_address[REAL_ADDR_BITS-1:0]}; 

wire [ADDRESS_BITS-1:0]  inetwork_address    = network2cache_address[(ADDRESS_BITS-1) -: ADDRESS_BITS]; 
wire [ADDRESS_BITS-1:0]  dnetwork_address    = network2cache_address[((2*ADDRESS_BITS)-1) -: ADDRESS_BITS];
wire [ADDRESS_BITS-1:0]  reform_i_network2cache_addr = {i_addr_dest, inetwork_address[REAL_ADDR_BITS-1:0]}; 
wire [ADDRESS_BITS-1:0]  reform_d_network2cache_addr = {d_destination, dnetwork_address[REAL_ADDR_BITS-1:0]}; 

wire [MSG_BITS-1:0]      i_mem2cache_msg     = i_local_addr ? m_mem2cache_msg[MSG_BITS-1:0]: network2cache_msg[MSG_BITS-1:0];
wire [ADDRESS_BITS-1:0]  i_mem2cache_address = i_local_addr ? reform_i_mem2cache_addr : reform_i_network2cache_addr;
wire [DATA_WIDTH-1:0]    i_mem2cache_data    = i_local_addr ? m_mem2cache_data[DATA_WIDTH-1:0]: network2cache_data[DATA_WIDTH-1:0];

wire [MSG_BITS-1:0]      d_mem2cache_msg     = d_local_addr ? m_mem2cache_msg[((2*MSG_BITS)-1) -:MSG_BITS]: 
															  network2cache_msg[((2*MSG_BITS)-1) -:MSG_BITS];
wire [ADDRESS_BITS-1:0]  d_mem2cache_address = d_local_addr ? reform_d_mem2cache_addr : reform_d_network2cache_addr; 
wire [DATA_WIDTH-1:0]    d_mem2cache_data    = d_local_addr ? m_mem2cache_data[((2*DATA_WIDTH)-1) -:DATA_WIDTH]: 
															  network2cache_data[((2*DATA_WIDTH)-1) -:DATA_WIDTH];

wire [(MSG_BITS * NUM_CACHES)-1:0]     mem2cache_msg     = {d_mem2cache_msg,i_mem2cache_msg}; 
wire [(ADDRESS_BITS * NUM_CACHES)-1:0] mem2cache_address = {d_mem2cache_address,i_mem2cache_address}; 
wire [(DATA_WIDTH * NUM_CACHES)-1:0]   mem2cache_data    = {d_mem2cache_data,i_mem2cache_data}; 

always @ (posedge clock) begin 
	active_dest <= reset? 0: i_destination; 
end 

unified_caches #(CORE, DATA_WIDTH, INDEX_BITS, OFFSET_BITS, ADDRESS_BITS, MSG_BITS) 
             unified (
				     clock, reset,
				       
				     read, write, address, in_data, 
				     out_address, out_data, valid, ready,
				     
				     mem2cache_msg, 
				     mem2cache_address, mem2cache_data, 
				     
				     cache2mem_msg, 
				     initial_cache2mem_addr, cache2mem_data, 
					 report
); 

main_memory #(CORE, DATA_WIDTH, ADDRESS_BITS, REAL_ADDR_BITS, M_CACHES, MSG_BITS, OFFSET_BITS, INIT_FILE) 
              m_memory (
				     clock, reset,  
				     				     
				     lm_cache2mem_msg, 
				     lm_cache2mem_address, lm_cache2mem_data,
				     
				     lm_mem2cache_msg, 
				     lm_mem2cache_address, lm_mem2cache_data
);


packetizer #(CORE, DATA_WIDTH, ADDRESS_BITS, NUM_CACHES, MSG_BITS, OFFSET_BITS, 
             IN_PORTS, OUT_PORTS, VC_BITS, 
             VC_DEPTH_BITS, ID_BITS, EXTRA, TYPE_BITS) network_packetizer (
				     clock, reset,  
				     				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data,
				     
				     mem2network_msg, 
				     mem2network_address, mem2network_data,
				     
				     network2cache_msg, 
				     network2cache_address, network2cache_data,
				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data, 
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full 
);

reg [31:0] remote_access; 
reg [31:0] remote_service; 

always @ (posedge clock) begin   
	if(reset) begin 
		remote_access  <= 0; 
		remote_service <= 0; 
	end 
	else begin 
		if(cache2network_msg[((2*MSG_BITS)-1) -: MSG_BITS] > C_RECV) remote_access  <= remote_access + 1; 
		if((network2mem_msg == R_REQ) | (network2mem_msg == W_REQ)) remote_service   <= remote_service + 1; 
	end 

 if (report) begin 
	  $display ("-------------------------- Memory back-end %d ----------------------------", Core_ID);  
	  $display ("Local cache remote accesses [%d]  Local memory remote services [%d]", remote_access, remote_service);
	  $display ("------------------------------------------------------------------------------");
	 end 
 end
	  
/*
 always @ (posedge clock) begin   
	$display ("--------------------------------System Wrapper Core %d ----------------------------------", Core_ID);  
	$display ("cache2mem_msg [%d]\t| cache2mem_address [%h]| cache2mem_data [%h]", cache2mem_msg, cache2mem_address,cache2mem_data ); 
	$display ("i_cache2mem_msg [%d]\t| i_cache2mem_address [%h]| i_cache2mem_data [%h]", i_cache2mem_msg, i_cache2mem_address,i_cache2mem_data );
	$display ("d_cache2mem_msg [%d]\t| d_cache2mem_address [%h]| d_cache2mem_data [%h]", d_cache2mem_msg, d_cache2mem_address,d_cache2mem_data );

	$display ("Reset [%b]", reset); 
	$display ("i_destination [%d]\t| i_local_addr [%b], i_addr_dest[%b] ", i_destination, i_local_addr,  i_addr_dest);
	$display ("d_destination [%d]\t| d_local_addr [%b] ", d_destination, d_local_addr);

	$display ("Init_I_address [%h]\t| iShort_cache2mem_address [%h]| iFull_cache2mem_address [%h]", i_initial_addr, iShort_cache2mem_address,iFull_cache2mem_address);
	$display ("Init_D_address [%h]\t| dShort_cache2mem_address [%h]| dFull_cache2mem_address [%h]", d_initial_addr, dShort_cache2mem_address,dFull_cache2mem_address);

	$display ("m_cache2mem_msg [%d]\t| m_cache2mem_address [%h]| m_cache2mem_data [%h]", m_cache2mem_msg, m_cache2mem_address,m_cache2mem_data );
	$display ("n_cache2mem_msg [%d]\t| n_cache2mem_address [%h]| n_cache2mem_data [%h]", n_cache2mem_msg, n_cache2mem_address,n_cache2mem_data );

	$display ("lm_cache2mem_msg [%h]| lm_cache2mem_address [%h]| lm_cache2mem_data [%h]", lm_cache2mem_msg, lm_cache2mem_address,lm_cache2mem_data );
	$display ("mem2network_msg [%d]\t| mem2network_address [%h]| mem2network_data [%h]", mem2network_msg, mem2network_address,mem2network_data );
	$display ("cache2network_msg [%d]| cache2network_address [%h]| cache2network_data [%h]", cache2network_msg, cache2network_address,cache2network_data );
	$display ("m_mem2cache_msg [%d]\t| m_mem2cache_address [%h]| m_mem2cache_data [%h]", m_mem2cache_msg, m_mem2cache_address,m_mem2cache_data );

	$display ("i_mem2cache_msg [%d]\t| i_mem2cache_address [%h]| i_mem2cache_data [%h]", i_mem2cache_msg, i_mem2cache_address,i_mem2cache_data );
	$display ("d_mem2cache_msg [%d]\t| d_mem2cache_address [%h]| d_mem2cache_data [%h]", d_mem2cache_msg, d_mem2cache_address,d_mem2cache_data );
	$display ("mem2cache_msg [%d]\t| mem2cache_address [%h]| mem2cache_data [%h]", mem2cache_msg, mem2cache_address,mem2cache_data );
	$display ("network2cache_msg [%d]| network2cache_address [%h]| network2cache_data [%h]", network2cache_msg, network2cache_address,network2cache_data );
	$display ("network2mem_msg [%d]\t| network2mem_address [%h]| network2mem_data [%h]", network2mem_msg, network2mem_address,network2mem_data );
	$display ("reform_d_mem2cache_addr [%h]| reform_i_mem2cache_addr [%h]", reform_d_mem2cache_addr, reform_i_mem2cache_addr); 
	$display ("reform_d_network2cache_addr [%h]| reform_i_network2cache_addr[%h]", reform_d_network2cache_addr, reform_i_network2cache_addr); 
	$display ("-----------------------------------------------------------------------------------"); 
 end
//*/



endmodule
