/** @module : Dir_cache_interface
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
module cache_interface #(parameter CORE = 0, DATA_WIDTH = 32, OFFSET_BITS = 3, ADDRESS_BITS = 20,  MSG_BITS = 4) (
				     clock, reset,  
				     //Cache
				     in_msg, in_address, in_data_line, 
				     out_msg, out_address, out_data_line,
				     
				     //Cache2Memory
				     mem2cache_msg, 
				     mem2cache_address, mem2cache_data, 
				     
				     cache2mem_msg, 
				     cache2mem_address, cache2mem_data
);

localparam OFFSET   = 1<<OFFSET_BITS; 
localparam BLOCK_BITS  = DATA_WIDTH * OFFSET; 

input clock, reset; 

input [MSG_BITS-1:0]      in_msg; 
input [ADDRESS_BITS-1:0]  in_address; 
input [BLOCK_BITS-1:0]    in_data_line;
 				     
output [MSG_BITS-1:0]     out_msg; 
output [ADDRESS_BITS-1:0] out_address; 
output [BLOCK_BITS-1:0]   out_data_line;

//Cache2Memory
input [MSG_BITS-1:0]      mem2cache_msg; 
input [ADDRESS_BITS-1:0]  mem2cache_address; 
input [DATA_WIDTH-1:0]    mem2cache_data;
 				     
output [MSG_BITS-1:0]     cache2mem_msg; 
output [ADDRESS_BITS-1:0] cache2mem_address; 
output [DATA_WIDTH-1:0]   cache2mem_data; 

localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses 

localparam IDLE = 0, READ_Q = 1, RECIEVING = 2, SENDING = 3, WAIT = 4, DONE = 5;
localparam VC_DEPTH_BITS = 3; 
localparam M_INFO_WIDTH = MSG_BITS + ADDRESS_BITS + DATA_WIDTH; 
localparam C_INFO_WIDTH = MSG_BITS + ADDRESS_BITS + BLOCK_BITS;
localparam NO_OP = 0, M_REP_OP = 1, M_REQ_OP = 2,  C_REP_OP = 3,
		   C_REQ_OP = 4, C_REP_OP1 = 5, C_REP_OP2 = 6; 
		   
reg [OFFSET_BITS :0] counter; 
reg [3:0] state;

wire [(DATA_WIDTH * OFFSET)-1:0]  new_data_block;
wire [DATA_WIDTH -1:0]send_words [0:OFFSET-1];
reg  [DATA_WIDTH -1:0]rec_words [0:OFFSET-1];

reg [MSG_BITS-1:0]     t_out_msg; 
reg [ADDRESS_BITS-1:0] t_out_address;
reg [BLOCK_BITS-1:0]   t_out_data_line; 

reg [MSG_BITS-1:0]    t_cache2mem_msg; 
reg [ADDRESS_BITS-1:0]t_cache2mem_address; 
reg [DATA_WIDTH-1:0]  t_cache2mem_data;

reg [ADDRESS_BITS-1:0]expected_mem2cac_addr; 
reg [MSG_BITS-1:0]    cur_mem2cache_msg; 
reg [ADDRESS_BITS-1:0]cur_mem2cache_addr; 

reg [ADDRESS_BITS-1:0]cur_cache2mem_address; 

reg [MSG_BITS-1:0]               cur_message; 
reg [ADDRESS_BITS-1:0]           cur_address; 
reg [(DATA_WIDTH * OFFSET)-1:0]  cur_data_block;
	
integer index; 
reg  c_req_read;
reg  c_resp_read;

wire c_req_valid;
wire [C_INFO_WIDTH-1: 0] c_req_data;
wire c_req_full;
wire c_req_empty;
  
wire c_resp_valid;
wire [C_INFO_WIDTH-1: 0] c_resp_data;
wire c_resp_full;
wire c_resp_empty;

wire c_req_write    = ((in_msg == M_REQ) | (in_msg == W_REQ));
wire [C_INFO_WIDTH-1: 0] c_req_in_data = {in_msg, in_address, in_data_line};

wire c_resp_write   = ((in_msg == INV_RESP) | (in_msg == WB_RESP) | (in_msg == SH_RESP) | (in_msg == WB_INV_RESP)); 
wire [C_INFO_WIDTH-1: 0] c_resp_in_data = {in_msg, in_address, in_data_line};

reg c_resp_ready, c_req_ready; 
reg [MSG_BITS-1:0]       c2m_rq_msg; 
reg [ADDRESS_BITS-1:0]   c2m_rq_address; 
reg [BLOCK_BITS-1:0]     c2m_rq_data; 

reg [MSG_BITS-1:0]       c2m_rp_msg; 
reg [ADDRESS_BITS-1:0]   c2m_rp_address; 
reg [BLOCK_BITS-1:0]     c2m_rp_data; 

reg  m_req_read;
reg  m_resp_read;

wire m_req_valid;
wire [M_INFO_WIDTH-1: 0] m_req_data;
wire m_req_full;
wire m_req_empty;
  
wire m_resp_valid;
wire [M_INFO_WIDTH-1: 0] m_resp_data;
wire m_resp_full;
wire m_resp_empty;

wire m_req_write    = ((mem2cache_msg == M_REQ) | (mem2cache_msg == INV_REQ) | 
					  (mem2cache_msg == M_WB_REQ) | (mem2cache_msg == SHARE_REQ) | 
					  (mem2cache_msg == WB_INV_REQ));
wire [M_INFO_WIDTH-1: 0] m_req_in_data = {mem2cache_msg, mem2cache_address, mem2cache_data};

wire m_resp_write   = ((mem2cache_msg == W_RESP) | (mem2cache_msg == EX_RESP) | (mem2cache_msg == SH_RESP)); 
wire [M_INFO_WIDTH-1: 0] m_resp_in_data = {mem2cache_msg, mem2cache_address, mem2cache_data};

reg m_resp_ready, m_req_ready; 
reg [MSG_BITS-1:0]       m2c_rq_msg; 
reg [ADDRESS_BITS-1:0]   m2c_rq_address; 
reg [DATA_WIDTH-1:0]     m2c_rq_data; 

reg [MSG_BITS-1:0]       m2c_rp_msg; 
reg [ADDRESS_BITS-1:0]   m2c_rp_address; 
reg [DATA_WIDTH-1:0]     m2c_rp_data; 

wire [2: 0] operation = reset? NO_OP : (~m_resp_empty)? M_REP_OP : 
					   (~c_resp_empty)? C_REP_OP : (~c_req_empty)? C_REQ_OP : 
					   (~m_req_empty)?  M_REQ_OP : NO_OP; 
wire execute = (m_resp_read  | m_req_read | c_resp_read  | c_req_read);					   
wire [2: 0] execute_oper = reset? NO_MSG : ((m2c_rp_msg == W_RESP) | (m2c_rp_msg == EX_RESP) | 
						   (m2c_rp_msg == SH_RESP))? M_REP_OP:  
						   ((c2m_rp_msg == WB_INV_RESP) | (c2m_rp_msg == WB_RESP))? C_REP_OP1:
						   ((c2m_rp_msg == INV_RESP) | (c2m_rp_msg == SH_RESP))?    C_REP_OP2: 
						   ((c2m_rq_msg == W_REQ) | (c2m_rq_msg == R_REQ))?         C_REQ_OP : NO_OP;  

genvar i;
generate   
	for (i = 0; i < OFFSET; i=i+1) begin : LN_OUT
      	assign send_words[i] = cur_data_block [(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH]; 
      	assign new_data_block [(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH] = rec_words[i];
  	end  
endgenerate
  
fifo  #(C_INFO_WIDTH, VC_DEPTH_BITS, 0) c_req_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(c_req_in_data),
	.wrtEn(c_req_write),
	.rdEn(c_req_read),
	.peek(1'b0),

	.read_data(c_req_data), 
	.valid(c_req_valid),			
	.full(c_req_full), 
	.empty(c_req_empty) 
);

fifo  #(C_INFO_WIDTH, VC_DEPTH_BITS, 0) c_resp_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(c_resp_in_data),
	.wrtEn(c_resp_write),
	.rdEn(c_resp_read),
	.peek(1'b0),

	.read_data(c_resp_data), 
	.valid(c_resp_valid),			
	.full(c_resp_full), 
	.empty(c_resp_empty) 
);
					
always @(posedge clock) begin 
	if(reset==1)begin 
		c2m_rq_msg     <= 0;
		c2m_rq_address <= 0;
		c2m_rq_data    <= 0; 

		c2m_rp_msg     <= 0;
		c2m_rp_address <= 0;
		c2m_rp_data    <= 0;
		
		c_req_read     <= 0; 
		c_resp_read    <= 0;
	end 
    else begin
		c_resp_read <= ((state == IDLE) & c_resp_ready & (~c_resp_empty) & (~c_resp_read))? 1: 0;  
		c_req_read  <= ((state == IDLE) & c_req_ready & c_resp_empty & (~c_req_empty) & (~c_req_read))? 1: 0; 
		
		c2m_rq_msg     <= c_req_valid? c_req_data[(C_INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		c2m_rq_address <= c_req_valid? c_req_data[((C_INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		c2m_rq_data    <= c_req_valid? c_req_data[((C_INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 

		c2m_rp_msg     <= c_resp_valid? c_resp_data[(C_INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		c2m_rp_address <= c_resp_valid? c_resp_data[((C_INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		c2m_rp_data    <= c_resp_valid? c_resp_data[((C_INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 
	end 
end

fifo  #(M_INFO_WIDTH, VC_DEPTH_BITS, 0) m_req_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(m_req_in_data),
	.wrtEn(m_req_write),
	.rdEn(m_req_read),
	.peek(1'b0),

	.read_data(m_req_data), 
	.valid(m_req_valid),			
	.full(m_req_full), 
	.empty(m_req_empty) 
);

fifo  #(M_INFO_WIDTH, VC_DEPTH_BITS, 0) m_resp_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(m_resp_in_data),
	.wrtEn(m_resp_write),
	.rdEn(m_resp_read),
	.peek(1'b0),

	.read_data(m_resp_data), 
	.valid(m_resp_valid),			
	.full(m_resp_full), 
	.empty(m_resp_empty) 
);
					
always @(posedge clock) begin 
	if(reset==1)begin 
		m2c_rq_msg     <= 0;
		m2c_rq_address <= 0;
		m2c_rq_data    <= 0; 

		m2c_rp_msg     <= 0;
		m2c_rp_address <= 0;
		m2c_rp_data    <= 0;
		
		m_req_read     <= 0; 
		m_resp_read    <= 0;
	end 
    else begin
		m_resp_read <= ((state == IDLE) & m_resp_ready &  (~m_resp_empty) & (~m_resp_read))? 1: 0; 
		m_req_read  <= ((state == IDLE) & m_req_ready & m_resp_empty & (~m_req_empty) & (~m_req_read))? 1: 0; 
		
		m2c_rq_msg     <= m_req_valid? m_req_data[(M_INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		m2c_rq_address <= m_req_valid? m_req_data[((M_INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		m2c_rq_data    <= m_req_valid? m_req_data[((M_INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 

		m2c_rp_msg     <= m_resp_valid? m_resp_data[(M_INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
		m2c_rp_address <= m_resp_valid? m_resp_data[((M_INFO_WIDTH - MSG_BITS)-1) -: ADDRESS_BITS] : 0; 
		m2c_rp_data    <= m_resp_valid? m_resp_data[((M_INFO_WIDTH - MSG_BITS - ADDRESS_BITS)-1) : 0] : 0; 
	end 
end

always @ (posedge clock) begin  
	if(reset == 1) begin 
		t_out_msg       <= NO_MSG; 
		t_out_address   <= 0;
		t_out_data_line <= 0;
	end 
	else begin 
		if((m2c_rq_msg == INV_REQ) | (m2c_rq_msg == M_WB_REQ) | 
		  (m2c_rq_msg == SHARE_REQ) | (m2c_rq_msg == WB_INV_REQ)) begin 
			t_out_msg       <= m2c_rq_msg; 
			t_out_address   <= m2c_rq_address;
			t_out_data_line <= 0;
		end 
		else begin 
			if ((state == RECIEVING) & (m2c_rq_msg == NO_MSG) & (counter == OFFSET)) begin 
				t_out_msg       <= cur_mem2cache_msg; 
				t_out_address   <= cur_mem2cache_addr;
				t_out_data_line <= new_data_block;
			end 
			else begin 
				t_out_msg       <= NO_MSG; 
				t_out_address   <= 0;
				t_out_data_line <= 0;
			end 
		end 
	end 
end 

always @ (posedge clock) begin  
  if(reset == 1) begin 
		t_cache2mem_msg       <= NO_MSG; 
		t_cache2mem_address   <= 0; 
		t_cache2mem_data      <= 0;
		
		expected_mem2cac_addr <= 0; 
		cur_mem2cache_msg     <= NO_MSG; 
		cur_mem2cache_addr    <= 0; 
		cur_cache2mem_address <= 0; 
		cur_data_block	      <= 0;
		
		for (index = 0; index < OFFSET; index=index+1) begin 
			rec_words[index]  <= 0; 
		end 
		counter               <= 0;
		m_resp_ready          <= 0;
		c_resp_ready          <= 0;
		c_req_ready           <= 0;
		m_req_ready           <= 0;
		state                 <= IDLE; 
  end 
  else begin 
         case (state)
          	IDLE: begin
				state         <= (execute)? READ_Q : IDLE;
				m_resp_ready  <= (operation == M_REP_OP)? 1 : 0;
				c_resp_ready  <= (operation == C_REP_OP)? 1 : 0;
				c_req_ready   <= (operation == C_REQ_OP)? 1 : 0;
				m_req_ready   <= (operation == M_REQ_OP)? 1 : 0;  
				
				t_cache2mem_msg       <= NO_MSG; 
				t_cache2mem_address   <= 0; 
				t_cache2mem_data      <= 0; 
				expected_mem2cac_addr <= 0; 
				cur_cache2mem_address <= 0; 
				cur_data_block	      <= 0;
				counter               <= 0; 
			end
			
			READ_Q: begin 
				case (execute_oper)
					M_REP_OP: begin
						t_cache2mem_msg       <= C_READY; 
						t_cache2mem_address   <= m2c_rp_address; 
						t_cache2mem_data      <= CORE;
						
						expected_mem2cac_addr <= m2c_rp_address; 
						cur_mem2cache_msg     <= m2c_rp_msg; 
						cur_mem2cache_addr    <= m2c_rp_address;
						counter               <= 0;
						state                 <= RECIEVING;
					end
					C_REP_OP1: begin
						t_cache2mem_msg       <= c2m_rp_msg; 
						t_cache2mem_address   <= c2m_rp_address;  
						t_cache2mem_data      <= CORE; 	
						cur_cache2mem_address <= c2m_rp_address;
						cur_data_block	      <= c2m_rp_data; 					
						counter               <= 0;
						state                 <= SENDING;
					end
					C_REP_OP2: begin
						t_cache2mem_msg       <= c2m_rp_msg; 
						t_cache2mem_address   <= c2m_rp_address; 
						t_cache2mem_data      <= CORE; 
						state                 <= IDLE;
					end
					C_REQ_OP: begin
						t_cache2mem_msg       <= c2m_rq_msg; 
						t_cache2mem_address   <= c2m_rq_address; 
						t_cache2mem_data      <= CORE; 
						state                 <= IDLE;
					end
					default: begin
						t_cache2mem_msg       <= NO_MSG; 
						t_cache2mem_address   <= 0; 
						t_cache2mem_data      <= 0; 
						expected_mem2cac_addr <= 0; 
						cur_cache2mem_address <= 0; 
						counter               <= 0;
						state                 <= IDLE;
					end
				endcase 
				m_resp_ready  <= 0;
				c_resp_ready  <= 0;
				c_req_ready   <= 0;
				m_req_ready   <= 0;
			end 

			RECIEVING: begin
                if ((mem2cache_msg == MEM_SENT) & (expected_mem2cac_addr == mem2cache_address)) begin
                	 if(counter < OFFSET) begin
                	    t_cache2mem_msg       <= C_RECV;
						rec_words[counter]    <= mem2cache_data; 
                	    counter               <= counter + 1;   
						expected_mem2cac_addr <= expected_mem2cac_addr + 1; 
                	  end
                	  else begin
                	    t_cache2mem_msg        <= NO_MSG;
                	  end
                 end
				 else begin 
					if (mem2cache_msg == NO_MSG) begin 
						state            <= (counter == OFFSET)? DONE : RECIEVING; 
						t_cache2mem_msg  <= NO_MSG;
					end 
				end 
            end
			
            SENDING: begin
                if ((mem2cache_msg == M_READY) & (mem2cache_address == cur_cache2mem_address)) begin      
	                if(counter < OFFSET) begin
	                   t_cache2mem_msg     <= C_SENT;
	                   t_cache2mem_data    <= send_words[counter]; 
	                   counter             <= counter + 1;
	                   state               <= SENDING;     
	                 end
                	 else begin
                	   t_cache2mem_msg     <= NO_MSG;
                	  state                <= DONE;
                	 end
                end
                else begin 
					t_cache2mem_msg       <= NO_MSG;
					state                 <= SENDING;     
				end 
             end 
			 
            DONE: begin
        		state             <= IDLE; 				
            end 
			
          	default: state <= IDLE;
         endcase    
	end
end
	
//----------------------------------------------------
// Drive the outputs
//----------------------------------------------------
assign  out_msg       = reset? 0: t_out_msg; 
assign  out_address   = reset? 0: t_out_address;
assign  out_data_line = reset? 0: t_out_data_line;

assign  cache2mem_msg     = reset? 0: t_cache2mem_msg; 
assign  cache2mem_address = reset? 0: t_cache2mem_address; 
assign  cache2mem_data    = reset? 0: t_cache2mem_data;
 
/*
 always @ (posedge clock) begin 
	$display ("------------------------ Cache Interface core [%d] -----------------------------", CORE);   
	$display ("IN_MSG [%d]\t\t| Address [%h]| Dataline [%h]", in_msg, in_address, in_data_line);
	$display ("OUT_MSG [%d]\t\t| Address [%h]| Dataline [%h]", out_msg, out_address, out_data_line);	
	$display (""); 
	$display ("M_resp_empty [%b]\t| M_req_empty [%b]| C_resp_empty [%b]| C_req_empty [%b]", m_resp_empty, m_req_empty, c_resp_empty, c_req_empty);
	$display ("Operation [%d]\t\t| Execute [%b]\t| Execute_oper [%d]", operation, execute, execute_oper); 
	$display (""); 
	$display ("M Rp-ready [%b]\t| M Rq-ready [%b]| C Rp-ready [%b]| C Rq-ready [%b]", m_resp_ready, m_req_ready, c_resp_ready, c_req_ready); 
	$display ("M Rp_write [%b]\t| Rp_read [%b]\t| Rq_write [%b]\t| Rq_read [%b]", m_resp_write, 
				m_resp_read, m_req_write, m_req_read); 
	$display ("M Req_valid [%b]\t| Req_data [%h]", m_req_valid, m_req_data); 
	$display ("M Resp_valid [%b]\t| Resp_data [%h]", m_resp_valid, m_resp_data); 
	$display ("C Rp_write [%b]\t| Rp_read [%b]\t| Rq_write [%b]\t| Rq_read [%b]", c_resp_write, 
				c_resp_read, c_req_write, c_req_read); 
	$display ("C Req_valid [%b]\t| Req_data [%h]",  c_req_valid, c_req_data); 
	$display ("C Resp_valid [%b]\t| Resp_data [%h]", c_resp_valid, c_resp_data); 
	$display ("C2M MSG [%d]\t\t| C2M Add [%h]| C2M DATA [%h]", c2m_rq_msg, c2m_rq_address, c2m_rq_data);
						
	$display (""); 
	$display ("State [%d]\t\t| Counter [%d]\t| Exp_address [%h] | Cur_address [%h]", state, counter, 
			  expected_mem2cac_addr, cur_cache2mem_address); 
	$display ("R_words [%h]\t| S_words [%h]", rec_words[counter], send_words[counter]);
	$display ("New_line [%h]", new_data_block);   
	$display ("Cur_data_block [%h]", cur_data_block);
	$display ("MEM2CACHE MSG [%d]\t| MEM2CACHE Add [%h]\t| MEM2CACHE DATA [%h]", 
				 mem2cache_msg, mem2cache_address, mem2cache_data);
	$display ("CACHE2MEM MSG [%d]\t| CACHE2MEM Add [%h]\t| CACHE2MEM DATA [%h]", 
				 cache2mem_msg, cache2mem_address, cache2mem_data);
	$display ("-----------------------------------------------------------------------------");
 end 	
//*/
endmodule

