/** @module : dir_direct_mapped_cache
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
module dir_direct_mapped_cache #(parameter CORE = 0, DATA_WIDTH = 32, INDEX_BITS = 6, 
                     OFFSET_BITS = 3, ADDRESS_BITS = 20) (
				     clock, reset,  
				     //Core2Cache
				     read, write, address, in_data, 
				     out_addr, out_data, valid, ready,
				     
				     //Cache2MemorySubSystem
				     mem2cache_share, 
					 mem2cache_req_share,
					 mem2cache_exclusive,
					 mem2cache_modify,
					 mem2cache_invalidate, 
					 mem2cache_writeback,
				     mem2cache_address, 
					 mem2cache_line, 
				     
				     cache2mem_read, 
					 cache2mem_write, 
					 cache2mem_invalid, 
					 cache2mem_share, 
					 cache2mem_writeback, 
					 cache2mem_address, 
				     cache2mem_line, 
					 // performance reporting
					 report
);
					 
localparam OFFSET   = 1<<OFFSET_BITS; 
localparam TAGLESS_BITS = INDEX_BITS + OFFSET_BITS; 
localparam TAG_BITS = ADDRESS_BITS - TAGLESS_BITS;

localparam STATUS_BITS = 4; 
localparam BLOCK_BITS  = DATA_WIDTH * OFFSET; 
localparam LINE_BITS   = STATUS_BITS + TAG_BITS + BLOCK_BITS; 

input clock, reset; 
//Core2Cache
input read, write;
input [ADDRESS_BITS-1:0] address;
input [DATA_WIDTH-1:0] in_data;
output valid, ready;
output[ADDRESS_BITS-1:0] out_addr;
output[DATA_WIDTH-1:0] out_data;

//Cache2MemorySubSystem
input mem2cache_share;
input mem2cache_req_share;
input mem2cache_exclusive;
input mem2cache_modify;
input mem2cache_invalidate;
input mem2cache_writeback;
input [ADDRESS_BITS-1:0] mem2cache_address;
input [BLOCK_BITS-1:0]  mem2cache_line;
				     
output cache2mem_read;
output cache2mem_write;
output cache2mem_invalid;
output cache2mem_share; 
output cache2mem_writeback;
output [ADDRESS_BITS-1:0] cache2mem_address;
output [BLOCK_BITS-1:0]   cache2mem_line; 
input  report;

localparam IDLE = 0, HIT_OR_MISS = 1, W_HIT_OR_MISS =  2, MISS = 3, 
           WRITE_BACK = 4, WB_WAIT = 5, READ_ST = 6, WRITE_ST = 7, WAIT = 8, 
           UPDATE = 9, UPDATE_DONE = 10, INVALIDATE = 11, INV_WAIT = 12;
reg [3:0] state;
wire [1:0] pad = 2'b00; 
wire [OFFSET_BITS-1 :0] zero_offset = 0;   

reg SG1_read, SG1_write;
reg [ADDRESS_BITS-1:0]  SG1_address;
reg [DATA_WIDTH-1:0]    SG1_in_data;

reg SG0_read, SG0_write;
reg [ADDRESS_BITS-1:0]  SG0_address;
reg [DATA_WIDTH-1:0]    SG0_in_data;

wire [LINE_BITS-1: 0] cur_line; 
wire [DATA_WIDTH -1:0] r_words [0:OFFSET-1];
wire [DATA_WIDTH -1:0] w_words [0:OFFSET-1];
wire [BLOCK_BITS-1:0]  new_data_block;
 
wire [OFFSET_BITS-1:0]  c2c_offset  = SG1_address [OFFSET_BITS-1:0];
wire [INDEX_BITS-1:0]   c2c_index   = SG1_address [TAGLESS_BITS-1:OFFSET_BITS];
wire [TAG_BITS-1:0]     c2c_tag     = SG1_address [ADDRESS_BITS-1: TAGLESS_BITS];

wire [STATUS_BITS-1:0] cur_status                = cur_line[(LINE_BITS-1) -: STATUS_BITS];
wire [TAG_BITS-1:0]    cur_tag                   = cur_line[ ((TAG_BITS + BLOCK_BITS) - 1) -: TAG_BITS];  
wire [BLOCK_BITS-1:0]  cur_data_block            = cur_line[BLOCK_BITS-1:0];

wire valid_bit   = cur_status[0]; 
wire dirty       = cur_status[1];
wire exclusive   = cur_status[2];
wire modify      = cur_status[3];

reg cache2mem_read;
reg cache2mem_write;
reg cache2mem_invalid;
reg cache2mem_share; 
reg cache2mem_writeback;
reg [ADDRESS_BITS-1:0] cache2mem_address;
reg [BLOCK_BITS-1:0]   cache2mem_line; 

reg [ADDRESS_BITS-1:0] write_back_addr; 
reg [BLOCK_BITS-1:0]   write_back_data; 
reg [ADDRESS_BITS-1:0] invalidate_addr; 

reg  prev_mem_cache_read; 
reg  t_mem2cache_invalidate, t_mem2cache_writeback; 
reg  t_mem2cache_share, t_mem2cache_exclusive, t_mem2cache_modify, t_mem2cache_req_share; 
reg [ADDRESS_BITS-1:0] t_mem2cache_address;
reg [BLOCK_BITS-1:0]   t_mem2cache_line;

wire [INDEX_BITS-1:0]   mem2cache_index = mem2cache_address [TAGLESS_BITS-1:OFFSET_BITS];
wire mem_cache_read     = (mem2cache_invalidate | mem2cache_writeback | mem2cache_share | 
						   mem2cache_req_share | mem2cache_exclusive | mem2cache_modify);
wire mem_answer         = (t_mem2cache_share | t_mem2cache_exclusive | t_mem2cache_modify);
wire [ADDRESS_BITS-1:0] line_start_addr = ((SG1_address >> OFFSET_BITS)<<OFFSET_BITS);  

wire [TAG_BITS-1:0]     prev_mem2cache_tag   = t_mem2cache_address [ADDRESS_BITS-1: TAGLESS_BITS];
wire [INDEX_BITS-1:0]   prev_mem2cache_index = t_mem2cache_address [TAGLESS_BITS-1:OFFSET_BITS];
wire new_valid = ((t_mem2cache_share | t_mem2cache_exclusive  | t_mem2cache_writeback | 
				   t_mem2cache_req_share | t_mem2cache_modify) & (~t_mem2cache_invalidate));

wire mem_address_hit = (prev_mem2cache_tag == cur_tag) & valid_bit; 			   
wire [STATUS_BITS-1:0] mem_line_status  = {t_mem2cache_modify, t_mem2cache_exclusive, 1'b0, new_valid};
wire [LINE_BITS-1: 0]  mem_new_line     = (mem_answer)?      {mem_line_status, prev_mem2cache_tag, t_mem2cache_line} : 
										  (mem_address_hit)? {mem_line_status, prev_mem2cache_tag, cur_data_block}: 0;
wire mem_write_line = prev_mem_cache_read & (mem_answer | mem_address_hit); 
wire cache_busy = (mem_cache_read | mem_write_line);									
										
wire hit   = ((SG1_write| SG1_read) & ((c2c_tag != cur_tag) | (~valid_bit) | 
			 (SG1_write & ~(modify | exclusive)) | cache_busy))? 0 : 1; 
wire stall = ((c2c_index == SG0_address [TAGLESS_BITS-1:OFFSET_BITS]) & 
			 (SG0_write | SG0_read) & SG1_write)? 1 : 0; 
reg mem_stall;
wire m_stall = ((~cache_busy) & (state == HIT_OR_MISS) & (~hit) & mem_stall); 
	 
wire [INDEX_BITS-1:0]   local_line_index = reset?  0: ((write | read) & ((state == IDLE) | 
										  ((state == HIT_OR_MISS) & hit & ~stall)))? address[TAGLESS_BITS-1:OFFSET_BITS] :
										  ((SG0_write | SG0_read) & (state == UPDATE_DONE)& ~stall)? 
										  SG0_address[TAGLESS_BITS-1:OFFSET_BITS]:
										  ((SG1_write | SG1_read) & ((state == UPDATE) | m_stall))?										  
										  SG1_address[TAGLESS_BITS-1:OFFSET_BITS]: 0; 
										  
wire local_line_read                     = reset?  0: ((write | read) & ((state == IDLE) | 
										  ((state == HIT_OR_MISS) & hit & ~stall)))? 1 :
										  ((SG0_write | SG0_read) & (state == UPDATE_DONE)& ~stall)? 1:
										  ((SG1_write | SG1_read) & ((state == UPDATE) | m_stall))?	1: 0;
										
wire [LINE_BITS-1: 0] local_new_line     =  (SG1_write & (c2c_tag == cur_tag) & valid_bit & (modify | exclusive))? 
										    {4'b1011, cur_tag, new_data_block} : 
										    ((state == HIT_OR_MISS) & (~hit) & (dirty & valid_bit))? 
										    cache_busy? 0 : {4'b0000, cur_tag, new_data_block} : 0; 
										
wire local_line_write                    =  ((SG1_write & (c2c_tag == cur_tag) & valid_bit & (modify | exclusive)))? 1 : 
											cache_busy? 0 : ((state == HIT_OR_MISS) & (~hit) & (dirty & valid_bit))? 1 : 0; 
											
wire [INDEX_BITS-1:0] local_write_index  =  ((SG1_write & (c2c_tag == cur_tag) & valid_bit & (modify | exclusive)))? c2c_index : 
											cache_busy? 0 : ((state == HIT_OR_MISS) & (~hit) & (dirty & valid_bit))? c2c_index : 0; 

wire read_line                      = reset?  0: mem_cache_read? 1: local_line_read; 
wire [INDEX_BITS-1:0] cur_index     = reset?  0: mem_cache_read? mem2cache_index : local_line_index; 

wire flushing                       = reset?  0: ((state == WRITE_BACK) | (state == INVALIDATE))? 1 : 0;	   
wire write_line                     = reset?  0: flushing? 1 : prev_mem_cache_read? mem_write_line : local_line_write; 													 
wire [INDEX_BITS-1:0] write_index   = reset?  0: flushing? c2c_index : prev_mem_cache_read? prev_mem2cache_index : local_write_index;  
wire [LINE_BITS-1: 0] new_line      = reset?  0: flushing? 0 : prev_mem_cache_read? mem_new_line : local_new_line;			

// to collect statistical data 
reg [31: 0] total_hits;  
reg [31: 0] total_misses; 
	
genvar i;
generate   
	for (i = 0; i < OFFSET; i=i+1) begin : LN_OUT
      	assign r_words[i] = cur_data_block [(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH]; 
		assign new_data_block [(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH] = w_words[i];
  	end  
	for (i = 0; i < OFFSET; i=i+1) begin : LN_OUT2
      	assign w_words[i] = (SG1_write & (c2c_tag == cur_tag) & valid_bit & (modify | exclusive) &(c2c_offset == i))? 
							SG1_in_data : (state == WAIT)? w_words[i] : r_words[i];
  	end  
endgenerate
							
BRAM #(LINE_BITS, INDEX_BITS) CACHE_RAM (
		.clock(clock),
    	.readEnable(read_line),
    	.readAddress(cur_index),
   		.readData(cur_line),
		
    	.writeEnable(write_line),
    	.writeAddress(write_index),
    	.writeData(new_line)
); 

always @ (posedge clock) begin  
	if(reset == 1) begin 
		prev_mem_cache_read     <= 0;
		t_mem2cache_invalidate  <= 0; 
		t_mem2cache_writeback   <= 0;
		t_mem2cache_share       <= 0; 
		t_mem2cache_req_share   <= 0;		
		t_mem2cache_exclusive   <= 0; 
		t_mem2cache_modify      <= 0; 
		t_mem2cache_address     <= 0;
		t_mem2cache_line		<= 0; 
	end 
	else begin 
		prev_mem_cache_read     <= mem_cache_read;
		t_mem2cache_invalidate  <= mem2cache_invalidate; 
		t_mem2cache_writeback   <= mem2cache_writeback;
		t_mem2cache_share       <= mem2cache_share; 
		t_mem2cache_req_share   <= mem2cache_req_share;		
		t_mem2cache_exclusive   <= mem2cache_exclusive; 
		t_mem2cache_modify      <= mem2cache_modify; 
		t_mem2cache_address     <= mem2cache_address;
		t_mem2cache_line		<= mem2cache_line; 
	end 
end 

always @ (posedge clock) begin  
	if(reset == 1) begin 
		SG0_read            <= 0; 
		SG0_write           <= 0; 
		SG0_address         <= 0;
		SG0_in_data         <= 0;
		
		SG1_read            <= 0; 
		SG1_write           <= 0; 
		SG1_address         <= 0;
		SG1_in_data         <= 0;
		
		cache2mem_read      <= 0;
		cache2mem_write     <= 0;
		cache2mem_invalid   <= 0;
		cache2mem_share     <= 0; 
		cache2mem_writeback <= 0; 
		cache2mem_address   <= 0;
		cache2mem_line      <= 0; 
		
		write_back_addr     <= 0;
		write_back_data     <= 0; 
		invalidate_addr     <= 0; 
		mem_stall           <= 0;
		
		total_hits          <= 0;
		total_misses        <= 0;

		state               <= IDLE; 
	end 
	else begin 
		if (cache_busy) begin 
			cache2mem_read      <= 0;
			cache2mem_write     <= 0;
			cache2mem_invalid   <= t_mem2cache_invalidate; 
			cache2mem_writeback <= (t_mem2cache_writeback & mem_address_hit)? 1 : 0;
			cache2mem_share     <= (t_mem2cache_share | (t_mem2cache_req_share & mem_address_hit))? 1: 0;  
			cache2mem_address   <= t_mem2cache_address;
			cache2mem_line      <= cur_data_block; 
			
			mem_stall           <= 1;
			state               <= (mem_answer & (line_start_addr == t_mem2cache_address) & (state == WAIT))? 
									UPDATE :  ((state == IDLE) & (read | write))? HIT_OR_MISS : state; 
			
			if(state == IDLE) begin 
				SG1_read          <= read; 
				SG1_write         <= write;
				SG1_address 	  <= address;
				SG1_in_data       <= in_data;
			end 
		end 
		else begin 
			cache2mem_read      <= (state == READ_ST)?    1 : 0;
			cache2mem_write     <= (state == WRITE_ST)?   1 : 0;
			cache2mem_invalid   <= flushing?  1 : 0;
			cache2mem_share     <= 0;
			cache2mem_writeback <= (state == WRITE_BACK)? 1 : 0; 
			cache2mem_address   <= ((state == READ_ST) | (state == WRITE_ST))? line_start_addr : 
								   (state == WRITE_BACK)? write_back_addr : 
								   (state == INVALIDATE)? invalidate_addr : 0;
			cache2mem_line      <= (state == WRITE_BACK)? write_back_data : 0;
			
			mem_stall           <= 0;
		  case (state)
			IDLE: begin
				SG1_read          <= read; 
				SG1_write         <= write;
				SG1_address 	  <= address;
				SG1_in_data       <= in_data;
				state             <= (~read && ~write)? IDLE : HIT_OR_MISS; 
			end
					
			HIT_OR_MISS: begin
				if (hit) begin 
					if(stall) begin 
						SG1_read          <= 0; 
						SG1_write         <= 0;
						SG1_in_data       <= 0;
						
						SG0_read          <= read; 
						SG0_write         <= write;
						SG0_address 	  <= address;
						SG0_in_data       <= in_data;
						
						state             <= UPDATE_DONE; 
					end 
					else begin 
						SG1_read          <= read; 
						SG1_write         <= write;
						SG1_address 	  <= address;
						SG1_in_data       <= in_data;
						state             <= (~read && ~write)? IDLE : HIT_OR_MISS; 
					end 	
					total_hits            <= total_hits + 1;
				end
				else begin
					if(mem_stall) begin 
						state             <= HIT_OR_MISS; 
					end 
					else begin 
						SG0_read          <= read; 
						SG0_write         <= write;
						SG0_address 	  <= address;
						SG0_in_data       <= in_data;
					
						write_back_addr  <= (dirty & valid_bit)? {pad, cur_tag, c2c_index, zero_offset} : 0;						
						write_back_data  <= (dirty & valid_bit)? cur_data_block : 0; 
						invalidate_addr  <= valid_bit? {pad, cur_tag, c2c_index, zero_offset} : 0;
						
						state <= (dirty & valid_bit)? WRITE_BACK : valid_bit? INVALIDATE: 
								  SG1_read? READ_ST : SG1_write? WRITE_ST: HIT_OR_MISS;
					end 
					total_misses         <= total_misses + 1;
				end 
			end
			
			READ_ST: begin
				state               <= WAIT;
			end
							   
			WRITE_ST: begin 
				state               <= WAIT; 
			end

			WAIT: begin
				state               <= WAIT;  
			end
			
			UPDATE: begin
				state               <= UPDATE_DONE; 				
			end 
			
			UPDATE_DONE: begin
				if(stall) begin 
					SG1_read          <= 0; 
					SG1_write         <= 0;
					SG1_in_data       <= 0; 
					state             <= UPDATE_DONE ; 
				end 
				else begin 
					SG1_read          <= SG0_read; 
					SG1_write         <= SG0_write;
					SG1_address 	  <= SG0_address;
					SG1_in_data       <= SG0_in_data;

					SG0_read          <= 0; 
					SG0_write         <= 0;
					SG0_address 	  <= 0;
					SG0_in_data       <= 0;
					
					state             <= (~SG0_read && ~SG0_write)? IDLE : HIT_OR_MISS;	
				end 
			end 
			
			WRITE_BACK: begin
				state               <= WB_WAIT;
			end 
			
			INVALIDATE: begin
				state               <= INV_WAIT;
			end
			
			WB_WAIT: begin
				state               <= SG1_read? READ_ST : WRITE_ST;  
			 end 
			 
			INV_WAIT: begin
				state               <= SG1_read? READ_ST : WRITE_ST;  
			end 
			 
			default: state <= IDLE;
		  endcase 
		end 
	end
end 

//----------------------------------------------------
// Drive the outputs
//----------------------------------------------------
assign  out_addr = SG1_address; 
assign  valid    = (hit & (SG1_read | SG1_write))? 1 : 0; 
assign  out_data = (hit & SG1_read)? r_words[c2c_offset]: 0;
assign  ready    = (reset | (state == IDLE) | ((state == HIT_OR_MISS) & hit & ~stall))? 1 : 0;

// Performance data
reg [31 : 0] cycles; 

always @ (posedge clock) begin
	if (reset) begin	
		cycles           <= 0;  
	end 
    else begin
		cycles           <= cycles + 1;
		if (report) begin 
			$display ("Cache %d: Total hits [%d]\t| Total misses [%d]", CORE, total_hits, total_misses);
		end 	
	end 
end
/*
	always @ (posedge clock) begin    
		$display ("--------------------------------- Cache [%d] --------------------------------------", CORE);   
		$display ("Reset [%b]\t\t| Read [%b]\t\t| Write [%b]\t\t",reset, read, write);
		$display ("New Addr [%h]\t| New In_data [%h]", address, in_data);
		$display ("Valid [%b]\t| Addr [%h]\t| data [%h]", valid, address, out_data);
		$display ("Cur_read [%b]\t\t| Cur_write [%b]\t\t| Ready [%b]", SG0_read, SG0_write, ready);
		$display ("Cur Addr [%h]\t| Cur In_data [%h]", SG0_address, SG0_in_data);
		$display ("SG1_Read [%b]\t\t| SG1_Write [%b]", SG1_read, SG1_write);
		$display ("SG1 Addr [%h]\t| SG1 In_data [%h]", SG1_address, SG1_in_data);
		$display ("Offset [%d]\t\t| Index [%h]\t\t| TAG [%h]\t\t", c2c_offset, c2c_index, c2c_tag);
		$display ("Valid Bit [%b]\t\t| Dirty Bit[%b]\t\t| Hit [%b]", valid_bit, dirty, hit); 
		$display ("Found TAG [%h]\t| Found DATA [%h]", cur_tag, r_words[c2c_offset]);		
		$display ("State [%d]\t\t|", state); 
		$display ("Write Index [%h]\t| Write_line [%b]",write_index, write_line);
		$display ("New_line [%h]", new_line);
		$display ("Cur_Index [%h]\t| Read_line [%b]", cur_index, read_line);
		$display ("Cur_line [%h]", cur_line); 
		$display ("Cache_Read [%b]\t| mem2cache_index [%h]\t| local_line_index [%h]",
				  mem_cache_read, mem2cache_index, local_line_index);
		$display (""); 
		$display ("Out_valid [%b]\t\t| Out_Addr [%h]\t| Out_data [%h]", valid, out_addr, out_data);
		$display ("Stall [%b]\t\t| Mem_stall [%b]\t\t| Cache_busy [%b]", stall, mem_stall, cache_busy);
		$display ("");
		$display ("Cache2mem_read[%b]\t| Cache2mem_write[%b]\t| Cache2mem_invalid[%b]", cache2mem_read, cache2mem_write, cache2mem_invalid); 		
		$display ("Cache2mem_share[%b]\t| Cache2mem_writeback[%b]", cache2mem_share, cache2mem_writeback); 		
		$display ("Cache2mem_address[%h]", cache2mem_address); 
		$display ("Cache2mem_line[%h]", cache2mem_line); 	
		$display ("Mem2cache_share[%b]\t| Mem2cache_exclusive[%b]| Mem2cache_modify[%b]", mem2cache_share,  mem2cache_exclusive, mem2cache_modify);
		$display ("Mem2cache_invalidate[%b]|Mem2cache_writeback[%b]| Mem2cache_req_share [%b]", mem2cache_invalidate, mem2cache_writeback, 
					mem2cache_req_share); 		
		$display ("Mem2cache_address[%h]", mem2cache_address); 
		$display ("Mem2cache_line[%h]", mem2cache_line); 	
		$display ("-----------------------------------------------------------------------------");
	end 	
//*/
 
endmodule

