/** @module : 7_Stage_MIPS_2_Threaded_Core
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2011   Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

// 2-Threaded 7-Stage MIPS Core, fully-passed, no branch delay slot... 
module MIPS_core #(parameter CORE = 0, ID_BITS = 4, LOCAL_ADDR_BITS = 16, 
                            DATA_WIDTH = 32, INDEX_BITS = 6, 
                            OFFSET_BITS = 3, FULL_ADDRESS_BITS = 32, INIT_FILE = "memory.mem",
                            IN_PORTS = 1, OUT_PORTS = 1, VC_BITS = 1, VC_DEPTH_BITS = 4,
                            MSG_BITS = 4, EXTRA  = 2, TYPE_BITS = 2 , SWITCH_TO_SWITCH  = 1, 
							RT_ALG = 0, ROW = 3, COLUMN = 3)(
  clock, 
  reset, 
  start,
  prog_address, 
  
  from_peripheral,
  from_peripheral_data, 
  from_peripheral_valid, 
  to_peripheral,
  to_peripheral_data, 
  to_peripheral_valid,
  
  ON, PROG, WEn, REn,
  router_ID, rt_flowID, rt_entry, 
  s_flits_in, s_valid_in, 
  s_flits_out, s_valid_out, 
  s_empty_in, s_full_in, 
  s_empty_out, s_full_out, 
  // performance and debugging 
  report
); 

//define the log2 function
function integer log2;
	input integer num;
	integer i, result;
	begin
		for (i = 0; 2 ** i < num; i = i + 1)
			result = i + 1;
		log2 = result;
	end
endfunction

//---------------  Network and IO Inputs/Outputs -- Up to for 4 peripherals
localparam  FLOW_BITS     = (2*ID_BITS) + EXTRA;  
localparam  PAY_WIDTH     = MSG_BITS + DATA_WIDTH;
localparam  FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + PAY_WIDTH;
localparam  REAL_ADDR_BITS= LOCAL_ADDR_BITS -2; //Word alignment
localparam  ADDRESS_BITS  = ID_BITS + REAL_ADDR_BITS; 

localparam  PORTS = OUT_PORTS + 4*(SWITCH_TO_SWITCH);
localparam  PORTS_BITS = log2(PORTS); 
localparam  RT_WIDTH = PORTS_BITS + VC_BITS + 1; 
localparam  VC_PER_PORTS = (1 << VC_BITS); 

input  clock, reset, start; 
input  [31:0]  prog_address; 

input  [1:0]   from_peripheral;
input  [31:0]  from_peripheral_data; 
input  from_peripheral_valid;
output  [1:0]   to_peripheral;
output  [31:0]  to_peripheral_data; 
output  to_peripheral_valid;
reg  [1:0]   to_peripheral;
reg  [31:0]  to_peripheral_data; 
reg  to_peripheral_valid;

input ON, PROG, WEn, REn;
input [ID_BITS - 1: 0]    router_ID;
input [FLOW_BITS - 1: 0]  rt_flowID; 
input [RT_WIDTH - 1: 0]   rt_entry;
     
input [(FLIT_WIDTH * (SWITCH_TO_SWITCH *4)) - 1: 0] s_flits_in;
input [(SWITCH_TO_SWITCH *4) - 1: 0]s_valid_in; 
input [((VC_PER_PORTS * (SWITCH_TO_SWITCH *4)) - 1): 0] s_empty_in; 
input [((VC_PER_PORTS * (SWITCH_TO_SWITCH *4)) - 1): 0] s_full_in;
   
output [(FLIT_WIDTH * (SWITCH_TO_SWITCH *4)) - 1: 0] s_flits_out; 
output [(SWITCH_TO_SWITCH *4) - 1: 0] s_valid_out;
output [((VC_PER_PORTS * (SWITCH_TO_SWITCH *4)) - 1): 0] s_empty_out;
output [((VC_PER_PORTS * (SWITCH_TO_SWITCH *4)) - 1): 0] s_full_out; 

input report; 

wire   branch, zero, write ;  
wire   [2:0] PCSource; 
wire   [3:0] ALU_Op;
reg    [3:0] ALU_control;  
				 
reg    fetch, fetch_IC, load, store, write_EX, write_MEM, write_WB ; 
reg    load_MEM2, store_MEM2, write_MEM2; 
reg    [31:0]  PC_plus4, r_inst_addr, inst_addr_IC, r_data_addr, in_data;
reg    [31:0]  MEM2_addr; 
wire   [31:0]  instruction, out_data, jumpAddr, targetReg, branchAddr;
wire   [31:0]  IC_out_addr, DC_out_addr; 
wire   [31:0]  ALU_result, operand_A,  operand_B, read2_data;
reg    [31:0]  op_A, op_B, MD, result; 

wire   [4:0]   ws_reg; 
reg    [4:0]   ws_reg_WB, ws_reg_EX, ws_reg_MEM, ws_reg_MEM2 ;
reg    [31:0]  IR_D, IR_EX, IR_MEM, IR_MEM2, MEM_Data, MEM2_Data, WB_Data; 
reg    [31:0]  PC_D, PC_EX, PC_MEM, PC_MEM2; //, PC_WB;
reg    [31:0]  branchAddr_EX, targetReg_EX, jumpAddr_EX;
reg    [2:0]   PCSource_EX; 

//Bypass logic
wire [4:0]   rs     = IR_D[25:21];
wire [4:0]   rt     = IR_D[20:16];
wire logic_r_type; 
wire zeroEXT_type;
wire [2:0]  bypass1 = (rs == 0)? 0 :((rs == ws_reg_EX) & (write_EX))? 1:
					  ((rs == ws_reg_MEM) & (write_MEM))? 2:
					  ((rs == ws_reg_MEM2) & (write_MEM2))? 3: 
					  ((rs == ws_reg_WB) & (write_WB))? 4: 0; 
wire [2:0]  bypass2 = (rt == 0)? 0 :((rt == ws_reg_EX) & (write_EX))? 1:
					  ((rt == ws_reg_MEM) & (write_MEM))? 2:
					  ((rt == ws_reg_MEM2) & (write_MEM2))? 3:
					  ((rt == ws_reg_WB) & (write_WB))? 4: 0;	
					  		   
wire [2:0]   ASource  = logic_r_type? bypass2 : bypass1; 
wire [2:0]   BSource  = (logic_r_type | zeroEXT_type)? 0 : bypass2;
wire [2:0]   MDSource = logic_r_type? 0 : bypass2;

wire lw_stall = ((IR_EX[31:26] == 6'b100011) & ((bypass1 == 1) | (bypass2 == 1))) ? 1 : 
				((IR_MEM[31:26] == 6'b100011) & ((bypass1 == 2) | (bypass2 == 2))) ? 1 : 0; 
wire kill = ((PCSource_EX == 1)| (PCSource_EX == 2) | ((PCSource_EX == 3)& zero))? 1: 0; // kill on branch and jump 

localparam IDLE = 0, CLEAR = 1, ONSET = 2, INIT = 3, KERNEL = 4, RUN = 5; 
reg [2:0] state;
reg  stall; 
wire IC_ready, DC_ready;
wire i_valid, d_valid; 

wire cur_dValid      = (reset)? 0 : (DC_out_addr == MEM2_addr)? d_valid : 0;
wire d_stall = ((load_MEM2 | store_MEM2) & (~cur_dValid));	 
wire i_stall = ((fetch_IC & ~i_valid) | (~IC_ready)); 
wire IRSrc   = ~stall;  

reg t_cur_iValid; 
reg [31:0] t_cur_inst; 
reg [31:0] t_PC0, t_PC1; 

wire cur_iValid      =(reset | (inst_addr_IC == 0))? 0 : 
                      (IC_out_addr == inst_addr_IC)? i_valid : t_cur_iValid; 
wire [31:0] cur_inst =(reset | (inst_addr_IC == 0))? 0 : 
                      (IC_out_addr == inst_addr_IC)? instruction : t_cur_inst;					  
wire stop = 0; //Not stopping for now    

wire [ADDRESS_BITS-1:0] INS_out_addr; 
wire [ADDRESS_BITS-1:0] DAT_out_addr; 
wire [32-ADDRESS_BITS-1 :0] pad = 0; 	
reg  [ID_BITS - 1: 0]   program_core;
wire [ADDRESS_BITS-1:0] inst_addr = {program_core, r_inst_addr[REAL_ADDR_BITS+1:2]};
wire [ADDRESS_BITS-1:0] data_addr = {r_data_addr[ADDRESS_BITS+1:2]};

wire  [ID_BITS - 1: 0] inst_core = inst_addr_IC[ADDRESS_BITS+1 -:ID_BITS]; 
assign IC_out_addr =  {pad, inst_core, INS_out_addr[REAL_ADDR_BITS-1:0],2'b00}; 
assign DC_out_addr =  {pad, DAT_out_addr[ADDRESS_BITS-1:0], 2'b00}; 				  
wire forward_report; 

wire [ID_BITS - 1:0] core_ID = CORE;
wire [(2*ID_BITS) - 1:0] hThread0 = (CORE * 2); 
reg [31:0] hThread0_PC ;  
reg v_hThread0 ;

reg [(2*ID_BITS) - 1:0] hThread1; 
reg [31:0] hThread1_PC; 
reg v_hThread1; 

// Testing Variables  
wire [31:0]  time_sharing = 1024;  // switch thread every time_sharing cycles
reg finish = 0; 

reg  [31:0]  sharing_counter; 								
reg  [(2*ID_BITS) - 1:0]  t_cur_thread_id ; // each thread in the system will have a unique ID
wire [(2*ID_BITS) - 1:0] cur_thread_id = reset? 0 : t_cur_thread_id; // each thread in the system has a unique ID
reg  no_thread, switchThread;					
wire cur_thread = (cur_thread_id==hThread0)? 0 : 1;   
wire flush = reset? 0: (((time_sharing == sharing_counter) | finish) & 
             (((~cur_thread) & v_hThread1) | (cur_thread & v_hThread0)));

reg   [31:0] cur_thread_PC;
wire  [31:0] no_thread_PC = (~v_hThread1 & ~v_hThread0)? 1 : 0; 
										
wire  [31:0] PC0 = (cur_thread== 0)? ((state != RUN)? hThread0_PC : 
						  (lw_stall | d_stall)? t_PC0 : 
		        		  (PCSource_EX == 0)? PC_plus4 : 
		        		  (PCSource_EX == 1)?  jumpAddr_EX : 
		        		  (PCSource_EX == 2)?  targetReg_EX : 
		        		  zero? (PC_EX + 4 + branchAddr_EX) : 
						  i_stall?  t_PC0: PC_plus4): t_PC0;			  
						  
wire  [31:0] PC1 = (cur_thread==1)? ((state != RUN)? hThread1_PC : 
						 (lw_stall | d_stall)? t_PC1 : 
		        		 (PCSource_EX == 0)? PC_plus4 : 
		        		 (PCSource_EX == 1)?  jumpAddr_EX : 
		        		 (PCSource_EX == 2)?  targetReg_EX : 
		        		 zero? (PC_EX + 4 + branchAddr_EX) : 
						 i_stall?  t_PC1: PC_plus4) : t_PC1; 

wire  [31:0] PC =  (reset | no_thread)? no_thread_PC : 
				   (state == CLEAR)? no_thread_PC : cur_thread? PC1: PC0;	
				   
wire clear_signal = (state == CLEAR); 
					 
memory_router_system #(CORE, ID_BITS, REAL_ADDR_BITS, DATA_WIDTH, INDEX_BITS, 
                            OFFSET_BITS, ADDRESS_BITS, INIT_FILE, IN_PORTS, 
                            OUT_PORTS, VC_BITS, VC_DEPTH_BITS,
                            MSG_BITS, EXTRA, TYPE_BITS, SWITCH_TO_SWITCH, 
							RT_ALG, ROW, COLUMN)  
					memory_sub_system (
                     .clock(clock), 
                     .reset(reset),
				     .i_read(fetch), 
				     .i_write(1'b0), 
				     .i_address(inst_addr), 
				     .i_in_data(0), 
					 .i_out_addr(INS_out_addr),
				     .i_out_data(instruction), 
				     .i_valid(i_valid), 
				     .i_ready(IC_ready),
				     
				     .d_read(load), 
				     .d_write(store), 
				     .d_address(data_addr), 
				     .d_in_data(in_data), 
					 .d_out_addr(DAT_out_addr),
				     .d_out_data(out_data), 
				     .d_valid(d_valid), 
				     .d_ready(DC_ready), 
				    
				     .ON(ON),   .PROG(PROG),
				     .WEn(WEn), .REn(REn),
				     .router_ID(router_ID), 
				     .rt_flowID(rt_flowID), 
				     .rt_entry(rt_entry), 

  				     .s_flits_in(s_flits_in),  
  				     .s_valid_in(s_valid_in), 
  					 .s_flits_out(s_flits_out), 
  					 .s_valid_out(s_valid_out), 
  					 .s_empty_in(s_empty_in), 
  					 .s_full_in(s_full_in), 
					 .s_empty_out(s_empty_out), 
					 .s_full_out(s_full_out), 

					 // performance and debugging 
					 .report(forward_report)
); 

ALU #(32) alu_unit(
		.ALU_Op(ALU_control), 
		.operand_A(op_A), 
		.operand_B(op_B), 
		.ALU_result(ALU_result), 
		.zero(zero)
); 

two_inst_decoder decoder (
	  .clock(clock), 
	  .reset(reset), 
	  .PC(PC_D[31:28]), 
	  .instruction(IR_D), 
	  .branch(branch), 
	  .PCSource(PCSource), 
	  .ALU_Op(ALU_Op), 
	  .operand_A(operand_A), 
	  .operand_B(operand_B), 
	  .write(write), 
	  .ws_reg(ws_reg),
	  .read2_data(read2_data),
	  .logic_r_type(logic_r_type), 
	  .zeroEXT_type(zeroEXT_type), 
	  .branchAddr(branchAddr), 
	  .jumpAddr(jumpAddr),
	  .targetReg(targetReg), 
	  .RF_write(write_WB), 
	  .RF_ws_reg(ws_reg_WB), 
	  .RF_write_data(WB_Data),
	  .EX_thread(cur_thread)
); 

always @ (posedge clock) begin
	if (reset) begin 
		fetch        <= 0;
		stall        <= 1;
		state        <= IDLE;
		r_inst_addr  <= 0; 
		inst_addr_IC <= 0;
		fetch_IC	 <= 0; 		

		t_cur_inst   <= 0; 
		t_cur_iValid <= 0; 
		t_PC0        <= 0; 
		t_PC1        <= 0;
		PC_D         <= 0;

		r_data_addr  <= 0; 
		in_data      <= 0;
		load         <= 0; 
		store        <= 0;
		load_MEM2    <= 0; 
		store_MEM2   <= 0;
		MEM2_addr    <= 0;				

		cur_thread_PC<= 0; 
		$display (" Reset .................................");  
	end 
	else begin 
		t_cur_inst   <= (d_stall | lw_stall)? cur_inst   : 0; 
		t_cur_iValid <= (d_stall | lw_stall)? cur_iValid : 0;
		t_PC0        <= PC0; 
		t_PC1        <= PC1;
		
	 case (state)
		IDLE: begin
			state <= (start)? CLEAR : IDLE; 
		end
			 
		CLEAR: begin
			IR_D       <= 0; IR_EX       <= 0;  IR_MEM    <= 0;
			write_EX   <= 0; write_MEM   <= 0;  write_WB  <= 0;
			ws_reg_EX  <= 0; ws_reg_MEM  <= 0;  ws_reg_WB <= 0;
			MEM_Data   <= 0; WB_Data     <= 0;  IR_MEM2   <= 0;
			write_MEM2 <= 0; ws_reg_MEM2 <= 0;  MEM2_Data <= 0;
			
			op_A             <= 0; op_B         <= 0;  MD          <= 0; 	
			branchAddr_EX    <= 0; jumpAddr_EX  <= 0;  ALU_control <= 0;
			targetReg_EX    <= 0;		     	
			PC_plus4         <= 0; r_inst_addr  <= 0; 
			inst_addr_IC     <= 0; PCSource_EX  <= 0;
			PC_D             <= 0; PC_EX        <= 0;  
			PC_MEM           <= 0; PC_MEM2      <= 0;  fetch      <= 0;
			load_MEM2        <= 0; store_MEM2   <= 0;  fetch_IC   <= 0;
			write_WB     	 <= 0; ws_reg_WB    <= 0;  WB_Data    <= 0;
			cur_thread_PC    <= 0; MEM2_addr    <= 0;
			state            <= no_thread? IDLE: ONSET;	
			$display (" Clearing  ................................."); 
		end
		
		ONSET: begin
			state            <= no_thread? IDLE: INIT;	
			$display (" Onset  ................................."); 
			
			if(v_hThread0 & (switchThread & cur_thread)) begin 
				$display ("| Core %d active hardware thread %d program counter [%h]", CORE, hThread0, hThread0_PC);
			end 
			if(v_hThread1 & (switchThread & (~cur_thread))) begin 
				$display ("| Core %d active hardware thread %d program counter [%h]", CORE, hThread0, hThread0_PC);
			end 
		end
		
		INIT: begin
			cur_thread_PC <= cur_thread? hThread1_PC : hThread0_PC;
			state         <= KERNEL;
		end
		
		KERNEL: begin
			stall         <= 0; 
			fetch         <= 1;
			PC_plus4      <= PC + 4; 
			r_inst_addr   <= PC; 
			inst_addr_IC  <= 0;
			state         <= (PC > 0)? RUN : KERNEL;
		end

		RUN: begin
			 // Fetch			 
			 PC_plus4      <= kill? PC + 4 : (lw_stall | d_stall | i_stall)? PC_plus4    : PC + 4;
			 r_inst_addr   <= kill? PC     : (lw_stall | d_stall | i_stall)? r_inst_addr : PC; 
			 
			 // ICache Read
			 fetch_IC     <=  fetch;
			 inst_addr_IC <=  (d_stall | lw_stall)? inst_addr_IC : kill? 0 : i_stall? inst_addr_IC : r_inst_addr;  
				  
			 // Decode  
			 IR_D         <= d_stall? IR_D : kill? 0: lw_stall? IR_D : IRSrc? cur_inst : 32'h00000000; 
			 PC_D         <= d_stall? PC_D : kill? 0: lw_stall? PC_D : inst_addr_IC;
			 op_A	      <= d_stall? op_A : kill? 0: lw_stall? op_A: (ASource == 0)? operand_A : 
									  (ASource == 1)? ALU_result : (ASource == 2)? MEM_Data : 
									  (ASource == 3)? (IR_MEM2[31:26] == 6'b100011)? out_data : MEM2_Data :
									  (ASource == 4)? WB_Data : 0;
			 op_B	      <= d_stall? op_B : kill? 0: lw_stall? op_B:(BSource == 0)? operand_B : 
									  (BSource == 1)? ALU_result : (BSource == 2)? MEM_Data : 
									  (BSource == 3)? (IR_MEM2[31:26] == 6'b100011)? out_data : MEM2_Data : 
									  (BSource == 4)? WB_Data : 0;
									  
			 MD           <= d_stall? MD : kill? 0: lw_stall? MD:(MDSource == 0)? read2_data : (MDSource == 1)? 
										ALU_result : (MDSource == 2)? MEM_Data : 
										(MDSource == 3)? (IR_MEM2[31:26] == 6'b100011)? 
										out_data : MEM2_Data :(MDSource == 4)? WB_Data : 0; 
							
			 branchAddr_EX    <= d_stall? branchAddr_EX : kill? 0: lw_stall? branchAddr_EX : branchAddr;
			 PCSource_EX      <= d_stall? PCSource_EX : kill? 0: lw_stall? PCSource_EX : PCSource; 
			 
			 targetReg_EX     <= d_stall? targetReg_EX : kill? 0: lw_stall? targetReg_EX : 
										  (ASource == 0)? targetReg : (ASource == 1)? ALU_result : 
										  (ASource == 2)? MEM_Data : (ASource == 3)? (IR_MEM2[31:26] == 6'b100011)? 
										  out_data : MEM2_Data : (ASource == 4)? WB_Data : 0;
									  
			 jumpAddr_EX      <= d_stall? jumpAddr_EX : kill? 0: lw_stall? jumpAddr_EX : jumpAddr;					
				 
			 // EXECUTE  
			 PC_EX         <= d_stall? PC_EX : kill? 0: PC_D;
			 IR_EX         <= d_stall? IR_EX : kill? 0: lw_stall? 0: IR_D ;
			 write_EX      <= d_stall? write_EX : kill? 0: lw_stall? 0: write ; 
			 ws_reg_EX     <= d_stall? ws_reg_EX : kill? 0: lw_stall? 0: ws_reg ;	     
			 ALU_control   <= d_stall? ALU_control : kill? 0: lw_stall? 0: ALU_Op; 
		   
			 // Memory 
			 PC_MEM       <= d_stall? PC_MEM: PC_EX ;
			 IR_MEM       <= d_stall? IR_MEM: IR_EX ;
			 write_MEM    <= d_stall? write_MEM: write_EX ; 
			 ws_reg_MEM   <= d_stall? ws_reg_MEM: ws_reg_EX ;
			 r_data_addr  <= d_stall? r_data_addr: ALU_result; 
			 in_data      <= d_stall? in_data: MD;            // store data
			 load         <= d_stall? load: (IR_EX[31:26] == 6'b100011)? 1 : 0; 
			 store        <= d_stall? store: (IR_EX[31:26] == 6'b101011)? 1 : 0; 
			 MEM_Data     <= d_stall? MEM_Data: (write_EX & (ws_reg_EX == 31))? (PC_EX + 4) : ALU_result;
			 
			// Memory  Out
			 PC_MEM2        <= d_stall? PC_MEM2    : PC_MEM;
			 IR_MEM2        <= d_stall? IR_MEM2    : IR_MEM; 
			 write_MEM2     <= d_stall? write_MEM2 : write_MEM ; 
			 ws_reg_MEM2    <= d_stall? ws_reg_MEM2: ws_reg_MEM ;
			 MEM2_Data      <= d_stall? MEM2_Data  : MEM_Data; 
			 load_MEM2	  	<= d_stall? load_MEM2  : load; 
			 store_MEM2	  	<= d_stall? store_MEM2 : store;
			 MEM2_addr	  	<= d_stall? MEM2_addr  : r_data_addr;
			 
			 // WriteBack
			 write_WB     <= d_stall? write_WB : write_MEM2 ; 
			 ws_reg_WB    <= d_stall? ws_reg_WB: ws_reg_MEM2 ;
			 WB_Data      <= d_stall? WB_Data  : (IR_MEM2[31:26] == 6'b100011)? out_data : MEM2_Data;
		
			 cur_thread_PC <= d_stall? cur_thread_PC : 
			 ((~d_valid) & (load_MEM2 | store_MEM2) & (PC_MEM2 > 0))? PC_MEM2 :  
			 (PC_MEM > 0)? PC_MEM: cur_thread_PC; 
			 
			 state        <= (stop)? IDLE :(flush)? CLEAR : RUN;           	
		end
	 endcase
	end
end



always @ (posedge clock) begin      
	 if (reset)  begin
		no_thread       <= 1; 
		switchThread    <= 0; 
		t_cur_thread_id <= 0; 			
		v_hThread0      <= 0;  
		hThread0_PC     <= 0; 
		hThread1        <= 0;
		v_hThread1      <= 0;  
		hThread1_PC     <= 0;
		program_core    <= 0; 			
		sharing_counter <= 0; 
	 end
	 else begin 
		 if ((state == IDLE) & start)  begin
			no_thread       <= 0; 
			switchThread    <= 0; 
			t_cur_thread_id <= hThread0; 			

			v_hThread0      <= 1;  
			hThread0_PC     <= prog_address; 
			
			hThread1        <= hThread0 + 1;
			v_hThread1      <= 1;  
			hThread1_PC     <= prog_address + 4;  		
			program_core    <= prog_address[ADDRESS_BITS+1 -:ID_BITS]; 
			sharing_counter <= 1; 
		 end
		 else begin 
			sharing_counter <=  (time_sharing == sharing_counter)? 1 : sharing_counter + 1; 
			switchThread    <= flush; 
			no_thread       <=  (~v_hThread0 & ~v_hThread1)? 1: 0; 
				
			t_cur_thread_id <= (switchThread)? ((cur_thread)? hThread0: hThread1) :
								((~v_hThread0 & v_hThread1 & (~cur_thread))? 
								hThread1 : ((v_hThread0 & ~v_hThread1 & cur_thread)?
								hThread0 : cur_thread_id));
								
			hThread0_PC     <= (v_hThread0 & ~cur_thread & ((state == RUN) | (state == KERNEL) | switchThread))? 
							   cur_thread_PC: hThread0_PC;
							   
			hThread1_PC     <= (v_hThread1 & cur_thread & ((state == RUN) | (state == KERNEL) | switchThread))? 
							   cur_thread_PC : hThread1_PC;
		 end 
	 end  
end

// Deburging 
reg [31:0] cycles         = 0; 
reg [31:0] prev_inst      = 0; 
reg [31:0] inst_counter0  = 0; 
reg [31:0] inst_counter1  = 0;
reg [31:0] prev_IR        = 0;

always @ (posedge clock) begin    
	cycles    <= cycles + 1; 
	prev_inst <= IR_MEM2;
	prev_IR   <= IR_D; 

	if((IR_MEM2 > 0) & (IR_MEM2 != prev_inst)) begin 
		if (cur_thread_id == hThread0) begin 
			inst_counter0 <= inst_counter0 + 1; 
			//$display ("| Core [%d] Thread [%d] Executed instruction [%h] Count [%d] PC [%h]", CORE, cur_thread_id, IR_MEM2, inst_counter0, PC_MEM2);
		end 
		else begin 
			inst_counter1 <= inst_counter1 + 1; 
			//$display ("| Core [%d] Thread [%d] Executed instruction [%h] Count [%d] PC [%h]", CORE, cur_thread_id, IR_MEM2, inst_counter1, PC_MEM2);
		end 
	end 
	
	if(v_hThread0 & (state == INIT) & (cur_thread_id == hThread0)) begin 
		$display ("| Core %d Hard Thread 0 ID [%d] start program counter [%h]", CORE, hThread0, hThread0_PC);
	end 
	
	if(v_hThread1 & (state == INIT) & (cur_thread_id == hThread1)) begin 
		$display ("| Core %d Hard Thread 1 ID [%d] start program counter [%h]", CORE, hThread1, hThread1_PC);
	end 
	 
/*	
		 $display ("------------------MIPS Core %d  Current cycle %d---------------", CORE, cycles); 
		 $display ("| Hard Thread [%b] program counter [%h]", hThread0, hThread0_PC);		 
		 $display ("| Hard Thread [%b] program counter [%h]", hThread1, hThread1_PC);
         $display ("| Cur_thread [%b]  Cur_Th_PC [%h]  ", cur_thread, cur_thread_PC);
		 $display ("| Counter [%d]  Flush [%b] ", sharing_counter, flush);	 
		 $display ("| ");
		 $display ("| PC [%h]", PC);
		 $display ("| Next PC [%h]", PC_plus4);
		 $display ("| PC Source [%h]", PCSource);
		 $display ("| PC Src EX [%h]", PCSource_EX);
		 $display ("| Current PC [%h]", r_inst_addr); 
		 $display ("| IC Ready [%b]", IC_ready);
		 $display ("| IC Stall [%b]", i_stall);
		 $display ("| IRSrc [%b]", IRSrc ); 
		 $display ("|\t\t\t| Fetch Inst [%h]", instruction);
		 $display ("|\t\t\t| i_valid [%b]", i_valid);
		 $display ("|\t\t\t| IC_out_addr [%h]", IC_out_addr);
		 $display ("|\t\t\t| Cur Inst [%h]", cur_inst);
		 $display ("|\t\t\t| Cur i_valid [%b]", cur_iValid);		 
		 $display ("|\t\t\t| IC Read Addr [%h]", inst_addr_IC);
		 $display ("|\t\t\t|\t\t\t| LW Stall [%d]", lw_stall);
		 $display ("|\t\t\t|\t\t\t| Read 1 [%h]", operand_A);
		 $display ("|\t\t\t|\t\t\t| Read 2 [%h]", operand_B);  
		 $display ("|\t\t\t|\t\t\t| RegRead2 [%h]", read2_data); 
		 $display ("|\t\t\t|\t\t\t| Branch [%h]", branchAddr);  
		 $display ("|\t\t\t|\t\t\t| Jump   [%h]", jumpAddr); 
		 $display ("|\t\t\t|\t\t\t| Jump Reg   [%h]", targetReg);  
		 $display ("|\t\t\t|\t\t\t| Source A [%d]", ASource);
		 $display ("|\t\t\t|\t\t\t| Source B [%d]", BSource);           
		 $display ("|\t\t\t|\t\t\t| ALU_control [%d]", ALU_control);
		 $display ("|\t\t\t|\t\t\t| Branch_EX [%h]", branchAddr_EX);
		 $display ("|\t\t\t|\t\t\t| Branch Taken [%b]",zero);
		 $display ("|\t\t\t|\t\t\t| Operand A [%h]",op_A);
		 $display ("|\t\t\t|\t\t\t| Operand B [%h]", op_B);
		 $display ("|\t\t\t|\t\t\t| MD [%h] ", MD);
		 $display ("|\t\t\t|\t\t\t|\t\t\t| write_EX [%b]\t\t| write_MEM [%b]\t\t| write_MEM2 [%b]\t| write_WB [%b]",write_EX, write_MEM, write_MEM2, write_WB);
		 $display ("|\t\t\t|\t\t\t|\t\t\t| ws_reg_EX [%d]\t| ws_reg_MEM [%d]\t| ws_reg_MEM2 [%d]\t| ws_reg_WB [%d]",ws_reg_EX, ws_reg_MEM, ws_reg_MEM2, ws_reg_WB);
		 $display ("|\t\t\t|\t\t\t| IR_D [%h]\t| IR_EX [%h]\t| IR_MEM [%h]\t| IR_MEM2 [%h]",IR_D, IR_EX, IR_MEM, IR_MEM2);
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t| load [%b]\t\t| MEM2_load [%b]", load, load_MEM2);
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t| store [%b]\t\t| MEM2_store [%b]", store, store_MEM2);
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t| Data_addr [%h]\t| MEM2_addr [%h]", r_data_addr, MEM2_addr);
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t| Data_in [%h]\t| MEM2_Data_in [%h]", data_in, MEM2_data_in);  
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t|\t\t\t| d_stall [%b]", d_stall); 
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t|\t\t\t| d_valid [%b]", d_valid); 
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t|\t\t\t| DC_out_addr [%h]", DC_out_addr); 
		 $display ("|\t\t\t|\t\t\t|\t\t\t|\t\t\t| MEM_Data [%h]\t| MEM2_Data [%h]\t| WB_Data [%h]",MEM_Data, MEM2_Data, WB_Data);
         $display ("----------------------------------------------------------------------------------------");   
//*/
end

//Register $s0-$s7 [16-23] are saved across a call ... Heracles [17-20] for final results
wire reserved_reg = (write_WB & ((ws_reg_WB >= 17) & (ws_reg_WB <= 20))); 
wire reg_reset    = (cur_thread_id == hThread1)? ((inst_counter1 < 25) & (WB_Data ==0)) : 
                    ((inst_counter0 < 25) & (WB_Data ==0)); // Initial reset

reg result_report; 
reg [31:0] performance_counter [1:0]; 
always @ (posedge clock) begin            
	if (reserved_reg & ~reg_reset)  begin
		to_peripheral       <= 0;
		to_peripheral_data  <= WB_Data ; 
		to_peripheral_valid <= 1;
		finish              <= 1; 
		$display ("-------------------------- CORE %d -----------------------------------", CORE);
		$display (" Result: Thread [%d] Register [%d] Value = %d", cur_thread_id, ws_reg_WB, WB_Data);
		if (cur_thread_id == hThread1) begin 
			$display (" Current cycles [%d] instructions executed [%d]", cycles, inst_counter1);
			performance_counter[1] <= cycles;
		end 
		else begin 
			$display (" Current cycles [%d] instructions executed [%d]", cycles, inst_counter0);
			performance_counter[0] <= cycles;
		end
		$display ("-------------------------------------------------------------------------------");
	end
	else begin 
		to_peripheral_valid <= 0;
		finish              <= 0; 
	end 
	
	if (report & start)  begin
		$display ("-------------------------- CORE %d -----------------------------------", CORE);
                if (cur_thread_id ==hThread1)
                    $display (" Current cycles [%d] instructions executed [%d]", cycles, inst_counter1);
                else $display (" Current cycles [%d] instructions executed [%d]", cycles, inst_counter0);
		$display ("-------------------------------------------------------------------------------");
	end
	
	result_report <= reset? 0 : (reserved_reg & ~reg_reset); //change this logic for multiple prints.
end

//Finish or report activation 
assign forward_report = ((report | result_report) & start); 

endmodule
